@@ -172,7 +172,7 @@ struct msgb *iofd_msgb_alloc(struct osmo_io_fd *iofd)
 {
 	uint16_t headroom = iofd->msgb_alloc.headroom;
 
-	OSMO_ASSERT(iofd->msgb_alloc.size < 0xffff - headroom);
+	OSMO_ASSERT(iofd->msgb_alloc.size <= 0xffff - headroom);
 	return msgb_alloc_headroom_c(iofd, iofd->msgb_alloc.size + headroom, headroom, "osmo_io_msgb");
 }
 
@@ -209,11 +209,8 @@ struct msgb *iofd_msgb_pending_or_alloc(struct osmo_io_fd *iofd)
  */
 int iofd_txqueue_enqueue(struct osmo_io_fd *iofd, struct iofd_msghdr *msghdr)
 {
-	if (iofd->tx_queue.current_length >= iofd->tx_queue.max_length) {
-		LOGPIO(iofd, LOGL_ERROR, "enqueueing message failed (queue full, %u msgs). Rejecting msgb\n",
-		       iofd->tx_queue.current_length);
+	if (iofd->tx_queue.current_length >= iofd->tx_queue.max_length)
 		return -ENOSPC;
-	}
 
 	llist_add_tail(&msghdr->list, &iofd->tx_queue.msg_queue);
 	iofd->tx_queue.current_length++;
@@ -343,16 +340,8 @@ void iofd_handle_segmented_read(struct osmo_io_fd *iofd, struct msgb *msg, int r
 	do {
 		pending = NULL;
 		res = iofd_handle_segmentation(iofd, msg, &pending);
-		if (res != IOFD_SEG_ACT_DEFER || rc < 0) {
+		if (res != IOFD_SEG_ACT_DEFER || rc < 0)
 			iofd->io_ops.read_cb(iofd, rc, msg);
-			/* The user could unregister/close the iofd during read_cb() above.
-			 * Once that's done, it doesn't expect to receive any more events,
-			 * so discard it: */
-			if (!IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED)) {
-				msgb_free(pending);
-				return;
-			}
-		}
 		if (res == IOFD_SEG_ACT_HANDLE_MORE)
 			msg = pending;
 	} while (res == IOFD_SEG_ACT_HANDLE_MORE);
@@ -523,6 +512,7 @@ int osmo_iofd_sendto_msgb(struct osmo_io_fd *iofd, struct msgb *msg, int sendto_
 	rc = iofd_txqueue_enqueue(iofd, msghdr);
 	if (rc < 0) {
 		iofd_msghdr_free(msghdr);
+		LOGPIO(iofd, LOGL_ERROR, "enqueueing message failed (%d). Rejecting msgb\n", rc);
 		return rc;
 	}
 
@@ -598,6 +588,7 @@ int osmo_iofd_sendmsg_msgb(struct osmo_io_fd *iofd, struct msgb *msg, int sendms
 	rc = iofd_txqueue_enqueue(iofd, msghdr);
 	if (rc < 0) {
 		iofd_msghdr_free(msghdr);
+		LOGPIO(iofd, LOGL_ERROR, "enqueueing message failed (%d). Rejecting msgb\n", rc);
 		return rc;
 	}
 
@@ -746,19 +737,9 @@ int osmo_iofd_register(struct osmo_io_fd *iofd, int fd)
 {
 	int rc = 0;
 
-	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED)) {
-		/* If re-registering same fd, handle as NO-OP.
-		 * And it is an even more explicit NO-OP
-		 * if the caller passed in -1. */
-		if (fd < 0 || fd == iofd->fd)
-			return 0;
-		/* New FD should go through unregister() first. */
-		return -ENOTSUP;
-	}
-
 	if (fd >= 0)
 		iofd->fd = fd;
-	if (iofd->fd < 0) {
+	else if (iofd->fd < 0) {
 		/* this might happen if both osmo_iofd_setup() and osmo_iofd_register() are called with -1 */
 		LOGPIO(iofd, LOGL_ERROR, "Cannot register io_fd using invalid fd == %d\n", iofd->fd);
 		return -EBADF;
@@ -769,8 +750,6 @@ int osmo_iofd_register(struct osmo_io_fd *iofd, int fd)
 		return rc;
 
 	IOFD_FLAG_UNSET(iofd, IOFD_FLAG_CLOSED);
-	IOFD_FLAG_SET(iofd, IOFD_FLAG_FD_REGISTERED);
-
 	if ((iofd->mode == OSMO_IO_FD_MODE_READ_WRITE && iofd->io_ops.read_cb) ||
 	    (iofd->mode == OSMO_IO_FD_MODE_RECVFROM_SENDTO && iofd->io_ops.recvfrom_cb) ||
 	    (iofd->mode == OSMO_IO_FD_MODE_RECVMSG_SENDMSG && iofd->io_ops.recvmsg_cb)) {
@@ -793,14 +772,7 @@ int osmo_iofd_register(struct osmo_io_fd *iofd, int fd)
  */
 int osmo_iofd_unregister(struct osmo_io_fd *iofd)
 {
-	int rc;
-
-	if (!IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED))
-		return 0;
-
-	rc = osmo_iofd_ops.unregister_fd(iofd);
-	IOFD_FLAG_UNSET(iofd, IOFD_FLAG_FD_REGISTERED);
-	return rc;
+	return osmo_iofd_ops.unregister_fd(iofd);
 }
 
 /*! Retrieve the number of messages pending in the transmit queue.
@@ -873,10 +845,6 @@ int osmo_iofd_close(struct osmo_io_fd *iofd)
 
 	iofd->pending = NULL;
 
-	osmo_iofd_ops.read_disable(iofd);
-	osmo_iofd_ops.write_disable(iofd);
-	osmo_iofd_unregister(iofd);
-
 	rc = osmo_iofd_ops.close(iofd);
 	iofd->fd = -1;
 	return rc;
